/* This code utilized the copywritten MIT Athena X widget set and other than
   that is under copyright by James Morris and John Grotzinger circa 1993.  
   Permissions to use etc are granted as per those in the X11 copyright.  
   No warranties or guarantees are made implicity or explicity, nor is any 
   responsiblity is taken for anything involving use of this software; no 
   claims are made as to its performance or any other aspect; 
   use at your own risk.  */ 

#include <unistd.h>

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/List.h>
#include <X11/Xaw/Viewport.h>


/*****************************************************************************

This contains all the code needed to use the HelpButton pseudowidget.  I call
it a pseudowidget because it doesn't define a new widget class; it just makes
a Command widget and attaches some things to it automatically.  The syntax is

String name;          name of the help button
Widget parent;        parent of the help button
String array[][2];    help array, or NULL --- see below
String file;          help file, or NULL --- see below
String childrenname;  suffix for children's names --- see below
ArgList arglist;      passed directly to create the Command widget
int numargs;          counts arglist

help_button = XjCreateManagedHelpButton(name, parent, array, file, 
                                        childrenname, arglist, numargs);

The widget returned is actually a Command widget; arglist and numargs are
passed onto it.  It is given children:  a popup shell, containing a Viewport,
containing a List, which is the menu of help entries; and (child of the List)
another popup shell, containing a Viewport, containing a Text widget in which
the help entries are displayed.  Their names, respectively, are:  
"help menu" OR "help menu:  [childrenname]", "lport", "list",
"help answers" OR "help answers:  [childrenname]", "text port", and "text."
If childrenname is neither NULL nor zero-length, it is used in the names of
the shells as indicated.

The help button is automatically given two callbacks, one which pops up the
menu shell when the button is pushed, and one which releases the memory used
to store the help entries (at creation) when the button is destroyed.  The
menu List is automatically given a callback for entry selection that pops up
the text shell and puts the text of the selected entry in the Text widget,
scrolling the text's viewport to the top.

The actual help material:  this is defined either in the array argument or the
file argument.  If the array is not NULL it is used; if it is the file is used
(if it's also NULL, there will be no help).  You ought to make the file NULL
even if you have a non-NULL array to make the call clear.  The help array
should be defined elsewhere in your code in the following format:

String myhelp[][2] = {
  {"first menu entry",
"This is what will appear in the Text widget when the first menu entry is\n\
selected.\n"},
   {"second menu entry", "appears when second entry selected"},
  ...
   {"last real menu entry", "appears when last real entry selected"},
   {"exit help", "exit help"}};  

{"exit help", "exit help"} MUST be the last pair.  Selection of the "exit help"
menu entry will pop down the help windows.  
Since the array provides its own memory, the help button doesn't allocate any
on its own with this method.

If the array argument is NULL, the calculate() function below is called to
create arrays of strings for the menu and text entries; this memory will be
freed when the button is destroyed.  The format of the file should be as
follows:


arbitrary notes in the beginning, as long as there aren't any quotation marks;
if you want to be sure the help file is present in the source code, you could
put comment marks around it and use it as a header file --- the comment marks
will be ignored by calculate() since the first is before any quotes and the
second after termination.

"first menu entry"
This will appear in the Text widget when the first menu entry is selected.
Note that there are no quote marks around this; it is delineated by the last
quote mark preceding and the first succeeding.

"second menu entry"
This will appear when the second entry is selected.  There wasn't any actual
need for the blank line in there, but it makes the file easier to read.  For
that matter, there's no need for a newline between the menu entry and the
text, but it makes things easier to read.

...

"exit help"
This MUST be present.  Anything after it, like the current lines were they to
be in the file, will be ignored, so it's safe to put * / here so that the file
can be included as a header file.


Of course, if you were going to force your own help file to be present, you 
might as well incorporate it as an array, so that it will be part of the
executable itself.  The file option is really useful only if you want the user
to be able to create or at least edit the help file themselves; say if the
users are a teacher and a class of students, and the teacher needs to be able
to change the help file for the students' benefit.

That's all there is to the HelpButton.  I tried a vaargs version, but I had
some problems with it, so I punted.

*****************************************************************************/


/* make it safe to include more than once */
#ifndef _XjHelpButton_defined
#define _XjHelpButton_defined

extern void xj_fake_click();

/* calculate puts into the index the positions and lengths of the entries in
the help file, and makes the list of topics; returns number of entries */

static int calculate(String sourcefile, String **addrmenu, long *ind[2])
{
  char s[100];
  int i, n;
  FILE *fp;

  n = 0;
  *addrmenu = (String *) NULL;
  ind[0] = ind[1] = (long *) NULL;
  if ((fp = fopen(sourcefile, "r")) == NULL) return 0;

  do {
    *addrmenu = (String *) 
      XtRealloc((XtPointer) *addrmenu, (n+1) * sizeof(String *));
    ind[0] = (long *) XtRealloc((XtPointer) ind[0], (n+1) * sizeof(long *));
    ind[1] = (long *) XtRealloc((XtPointer) ind[1], (n+1) * sizeof(long *));
    ind[0][n] = (n ? ind[0][n-1] : 0);
    ind[1][n] = -1;

    do {
      fscanf(fp, "%c", s);
      if (n) ind[1][n-1]++;
      ind[0][n]++;
    } while (s[0] != '\"');

    i = 0;
    do
      ind[0][n] += fscanf(fp, "%c", s+i);
    while (s[i++] != '\"');
    s[i-1] = '\0';
    (*addrmenu)[n] = XtNewString(s);

  } while (strcmp((*addrmenu)[n++], "exit help"));

  fclose(fp);

  *addrmenu = (String *) 
    XtRealloc((XtPointer) *addrmenu, (n+1) * sizeof(String *));
  (*addrmenu)[n] = NULL;
  return n;
}    

/*****************************************************************************/

/* showlist is a callback that brings up the viewport with the list of topics*/

static void showlist(Widget widget, XtPointer clientData, XtPointer callData)
{
  Widget w = XtNameToWidget(widget, "help menu");
  Display *disp = XtDisplay(w);
  Window win;

  XtPopup(w, XtGrabNone);
  if ((win = XtWindow(XtNameToWidget(widget, "*help answers"))))
    XRaiseWindow(disp, win); 
  XRaiseWindow(disp, XtWindow(w)); 
}

/*****************************************************************************/

struct xjhelpbuttonmemorystruct {
  String *menu;
  String *entry;
  long *ind[2];
};

/****************************************************************************/

/* showtext is a callback that reads the help entry from the help file using
the index for position and length and puts the entry in the viewport for text,
which it pops up, unless "exit help" was selected, in which case it pops down
both viewports */

static void showtext(Widget widget, XtPointer clientData, XtPointer callData)
{
  struct xjhelpbuttonmemorystruct *memory = 
    (struct xjhelpbuttonmemorystruct *) clientData;
  int n = XawListShowCurrent(widget)->list_index;
  int stay;
  String buf, sourcefile;
  Widget w;
  FILE *fp;

  if (memory->entry)  /* help info in array */
    stay = strcmp(buf = memory->entry[n], "exit help");
  else   /* help info in file */
    if ((stay = (memory->ind[1][n] > 0))) {
      sourcefile = XtName(widget);
      fp = fopen(sourcefile, "r");
      fseek(fp, memory->ind[0][n], SEEK_SET);
      buf = XtCalloc(memory->ind[1][n] + 1, sizeof(char));
      fread(buf, 1, memory->ind[1][n], fp);
      buf[memory->ind[1][n]] = '\0';
      fclose(fp);
    }

  if (stay) {
    XtPopup(w = XtNameToWidget(widget, "help answers"), XtGrabNone);
    XRaiseWindow(XtDisplay(w), XtWindow(w));
    XtVaSetValues(XtNameToWidget(widget, "*text"), XtNstring, buf, NULL);
    
    if (!memory->entry) XtFree(buf);
    
    /* scroll the viewport to the top */
    xj_fake_click(XtNameToWidget(widget, "help answers.text port.vertical"), 
		  2); 
  }
  else {
    XawListUnhighlight(widget);
    XtPopdown(XtNameToWidget(widget, "help answers"));
    XtPopdown(XtParent(XtParent(widget)));
  }
}

/*****************************************************************************/

/* free the allocated memory */

static void free_helpbutton_memory(Widget widget, 
				   XtPointer clientData, XtPointer callData)
{
  struct xjhelpbuttonmemorystruct *memory = 
    (struct xjhelpbuttonmemorystruct *) clientData;
  
  XtFree((XtPointer) memory->menu);
  
  if (memory->entry)
    XtFree((XtPointer) memory->entry);
  else {
    XtFree((XtPointer) memory->ind[0]);
    XtFree((XtPointer) memory->ind[1]);
  }
  
  XtFree((XtPointer) memory);
}

/*****************************************************************************/

/* this of course creates the actual button and calls the other routines */

Widget XjCreateManagedHelpButton(String name, Widget parent, String array[][2],
				 String file, String childrenname,
				 ArgList arglist, int numargs)
{
  Widget button, listshell, listport, list, textshell, textport, text;
  char title[100];
  int num_entries, i;
  struct xjhelpbuttonmemorystruct *memory;

  memory = (struct xjhelpbuttonmemorystruct *) 
    XtCalloc(1, sizeof(struct xjhelpbuttonmemorystruct));  
  
  button = XtCreateManagedWidget(name, commandWidgetClass, parent, 
				 arglist, numargs); 
  
  if (array) {
    for (i = 0; strcmp(array[i][0], "exit help"); ++i);
    num_entries = i + 1;
    memory->menu = (String *) XtCalloc(num_entries, sizeof(String));
    memory->entry = (String *) XtCalloc(num_entries, sizeof(String));
    for (i = 0; i < num_entries; ++i) {
      memory->menu[i] = XtNewString(array[i][0]);
      memory->entry[i] = XtNewString(array[i][1]);
    }
  }
  else if (file) {
    num_entries = calculate(file, &(memory->menu), memory->ind);
    memory->entry = NULL;
  }  
  else num_entries = 0;

  if (num_entries) {
    XtAddCallback(button, XtNcallback, showlist, NULL);
    XtAddCallback(button, XtNdestroyCallback, free_helpbutton_memory, memory);

    if (childrenname && strlen(childrenname))
      strcat(strcpy(title, "help menu:  "), childrenname);
    else
      strcpy(title, "help menu");
    listshell = XtVaCreatePopupShell("help menu", wmShellWidgetClass, button,
				     XtNtitle, title, XtNinput, True, NULL);

    listport = XtVaCreateManagedWidget
      ("lport", viewportWidgetClass, listshell,
       XtNheight, (Dimension) ((num_entries > 30 ? 30 : num_entries) * 15 + 4),
       XtNallowVert, True, 
       NULL);

    list = XtVaCreateManagedWidget(array ? "list" : file, listWidgetClass, 
				   listport,
				   XtNdefaultColumns, 1,
				   XtNforceColumns, True,
				   XtNlist, memory->menu,
				   XtNnumberStrings, num_entries,
				   NULL);
    XtAddCallback(list, XtNcallback, showtext, memory);

    if (childrenname && strlen(childrenname))
      strcat(strcpy(title, "help answers:  "), childrenname);
    else
      strcpy(title, "help answers");
    textshell = XtVaCreatePopupShell("help answers", wmShellWidgetClass, list,
				     XtNtitle, title, XtNinput, True, NULL);

    textport = XtVaCreateManagedWidget("text port", viewportWidgetClass,
				       textshell, XtNheight, (Dimension) 300,
				       XtNallowVert, True,
				       NULL);

    text = XtVaCreateManagedWidget("text", asciiTextWidgetClass, textport,
				   XtNdisplayCaret, False,
				   XtNwidth, (Dimension) 580,
				   XtNheight, (Dimension) 1000,
				   NULL);
  }
  else {
    XtFree((XtPointer) memory);
    XtSetSensitive(button, False);
  }  

  return button;
}

#endif  /* _XjHelpButton_defined */




#ifndef _xj_fake_click_defined
#define _xj_fake_click_defined

/* send a fake button # button ButtonPress and ButtonRelease to 
   0, 0 on a widget */
void xj_fake_click(Widget widget, int button)
{
  XEvent event;
  Window junk;

  event.type = ButtonPress;
  event.xany.serial = (unsigned long) NULL;  
  event.xany.send_event = True;
  event.xany.display = XtDisplay(widget);
  event.xbutton.window = XtWindow(widget);
  event.xbutton.root = XRootWindow(event.xany.display,
                                   XDefaultScreen(event.xany.display));
  event.xbutton.subwindow = None;
  event.xbutton.time = XtLastTimestampProcessed(event.xany.display);
  event.xbutton.x = 0;
  event.xbutton.y = 0;
  XTranslateCoordinates(event.xany.display, event.xbutton.window,
                        event.xbutton.root, event.xbutton.x, event.xbutton.y,
                        &event.xbutton.x_root, &event.xbutton.y_root, &junk);
  event.xbutton.state = 0;
  event.xbutton.button = button;
  event.xbutton.same_screen = True;
  XtDispatchEvent(&event);

  event.type = ButtonRelease;
  event.xbutton.time = XtLastTimestampProcessed(event.xany.display);
  XtDispatchEvent(&event);

  return;
}
#endif  /* _xj_fake_click_defined */







