

/* Copyright (C) 1995 P. B. Flemings, J. P. Grotzinger, J. E. Morris
   see strata.h                              */
#include "strata.h"
#include "license.h"

#include "globals.h"
#include <math.h>


/* Read in the data output by the simulator.  

   Crucialspecs() reads in the file <library>dat.out, which has only a few 
   numbers in it and is output at the start of the simulation run; these 
   numbers are needed to even set up the plotter.  The function is distinct 
   because during a movie these numbers are needed at the beginning when the 
   others are not yet known.
   
   Readspecs() calls crucialspecs() (since that data is constant it doesn't
   matter if it's read more than once; also there's no memory allocation
   involved there), then reads the <library>s.out file for the sealevel and
   shoreboundary data.  It allocates space for the sealevel, shoreboundary,
   and timeline arrays.

   Freespecs() frees the memory allocated by readspecs().  This is needed 
   during movie runs, when readspecs() is called many times; readspecs()
   checks the need_to_free variable to see if it should call freespecs()
   first, then sets need_to_free to True after allocating.
   Also, the exit routine goodbye() calls freespecs() just to be nice.  

   Spec_name() is a convenience function that returns a string consisted of
   the library name with the function's single argument appended.  */

extern double kmarine, knonmarine; /* for boom*.c (seismic stuff) */
extern String spec_name(char suf[]);

extern Boolean havedata[SIMDATA];
extern int nhavedata;

int nnodes;
int *shoreboundary;
int timeloop[NSLICES];
double cut[NSLICES];
double *sealevel;
double **timeline[NSLICES][SIMDATA];
double simversion;
int modelnum;
double delta, timedelt;
char isoname[100][100], isounits[100][100];
int nisotopes;
double *isotope[100];

void crucialspecs(void)
{
  FILE *fp;
  int i, n;

  if (!(fp = fopen(spec_name("dat.out"), "r"))) {
    printf("file %s not found:  exiting\n", spec_name("dat.out"));
    exit(1);
  }

  fscanf(fp, "%d %lf", &modelnum, &simversion);
  /* skip thrust */
  fscanf(fp, "%lf %d %*d %d %lf", &delta, &nnodes, &tlstep, &timedelt);
  for (i = 0; i < NSLICES; ++i) fscanf(fp, "%lf", cut+i);
  fscanf(fp, "%d", &compaction);
  fscanf(fp, "%lf %lf", &knonmarine, &kmarine);
  fscanf(fp, "%lf %lf", &compcut, &compdecay);
  fscanf(fp, "%lf %lf %lf", &condsand, &condshale, &condfluid);
  if (simversion > 2.13) {
    fscanf(fp, "%d", &nisotopes);
    if (simversion > 2.131) 
      for (n = 0; n < nisotopes; ++n)
	fscanf(fp, "%s %s", isoname[n], isounits[n]);
    else {
      printf("Fatal error:  2.131 simbasin incompatible with %g plotbasin\n",
	     STRATA_VERSION);
      exit(1);
    }
  }
  else
    nisotopes = 0;
  fclose(fp);
  
  havedata[0] = True;        /* elevation */
  havedata[1] = compaction;  /* porosity */
  havedata[2] = True;  /* water depth / facies */
  havedata[3] = True;  /* max space (accum) */
  havedata[4] = True;  /* diff const */
  havedata[5] = condsand && condshale && condfluid;  /* temperature */
  havedata[6] = condsand && condshale && condfluid;  /* TTI */
  nhavedata = 0;
  for (i = 0; i < SIMDATA; ++i)
    if (havedata[i])
      ++nhavedata;
}


Boolean need_to_free = False;
extern double seascale, isoscale[100];

void readspecs(void)
{               
  int i, j, k;  
  double x;
  FILE *fp;
  void freespecs();  /* below */

  if (need_to_free) freespecs();

  crucialspecs();

  if (!(fp = fopen(spec_name("s.out"), "r"))) {
    printf("file %s not found:  exiting\n", spec_name("s.out"));
    exit(1);
  }
  for (i = 0; i < NSLICES; ++i) fscanf(fp, "%d", timeloop+i);

  shoreboundary = (int *) calloc(timeloop[NSLICES - 1], sizeof(int));  
  for (i = 0; i < timeloop[NSLICES - 1]; ++i)
    fscanf(fp, "%d", shoreboundary+i); 

  sealevel = (double *) calloc(timeloop[NSLICES - 1], sizeof(double));
  x = 0;
  for (i = 0; i < timeloop[NSLICES - 1]; ++i) {
    fscanf(fp, "%lf", sealevel+i);  
    if (sealevel[i] > x)
      x = sealevel[i];
    else if (-sealevel[i] > x)
      x = -sealevel[i];
  }
  seascale = x ? pow(10.0, ceil(log10(x))) : 1;

  for (k = 0; k < nisotopes; ++k) {
    isotope[k] = (double *) calloc(timeloop[NSLICES - 1], sizeof(double));
    x = 0;
    for (i = 0; i < timeloop[NSLICES - 1]; ++i) {
      fscanf(fp, "%lf", isotope[k]+i);
      if (isotope[k][i] > x)
	x = isotope[k][i];
      else if (-isotope[k][i] > x)
	x = -isotope[k][i];
    }
    isoscale[k] = x ? pow(10.0, ceil(log10(x))) : 1;
  }

  for (i = 0; i < NSLICES; ++i)
    for (j = 0; j < SIMDATA; ++j) {
      timeline[i][j] = (double**)calloc(timeloop[NSLICES -1], sizeof(double*));
      for (k = 0; k < timeloop[NSLICES - 1]; ++k)
        timeline[i][j][k] = (double *) calloc(nnodes, sizeof(double));
    } 

  fclose(fp);
  need_to_free = True;
}


void freespecs(void)
{
  int i, j, k;

  for (i = 0; i < NSLICES; ++i)
    for (j = 0; j < SIMDATA; ++j) {
      for (k = 0; k < timeloop[NSLICES - 1]; ++k)
	free(timeline[i][j][k]);
      free(timeline[i][j]);
    }
  
  free(shoreboundary);
  free(sealevel);
  for (i = 0; i < nisotopes; ++i)
     free(isotope[i]);
}      


/* return string with (last) appended to (library), ie spec_name("1.out") 
   is "run041.out" if library is "run04" */

static char spec_name_return_string[1000];
String spec_name(char last[])
{
  strcpy(spec_name_return_string, library);
  strcat(spec_name_return_string, last);

  return spec_name_return_string;
}
